/*
 *	$Id: HZinMthd.c,v 1.4 1992/01/01 06:24:08 ygz Exp ygz $
 */

/***********************************************************
Copyright 1991 by Yongguang Zhang.  All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of the authors not
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

THE AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/

#ifdef HANZI

#include "ptyx.h"		/* X headers included here. */
#include <X11/Xos.h>
#include <stdio.h>
#include <ctype.h>
#include <pwd.h>
#include <sys/param.h>

#include "data.h"
#include "error.h"

#ifndef X_NOT_STDC_ENV
#include <stdlib.h>
#else
extern char *calloc();
#endif
#if defined(macII) && !defined(__STDC__)  /* stdlib.h fails to define these */
char *calloc();
#endif /* macII */

#include "HZinput.h"

/*** Only these variables and functions are to be exported ***/

extern HZinputTable *cHZtbl;
extern int (*HZfilterInput)();
extern Char dpyInbuf[];
extern int dpyInbufLen;
extern Char dpyChoices[];
extern int dpyChoicesLen;
extern SetHZMode();
extern HZInitInputTable();
extern HZClearInputTable();
extern RemakeHZInput();
extern void HandleSwitchHZMode();
extern void switchHZmodeByNum();
extern int HZParseInput();
extern int HZencode();
                                 /*added by zxk for all new input function */ 
extern int showlxDic();
extern int getLXState();
extern int resetLX();
extern int lxselect();
extern int updateGBLXDic();
extern int wildKeyInput();
extern int wildKeyNext();
extern int wildKeyPrev();
extern int wildKeyProcess();
extern int WildKeyState();
extern int firstMatchKey();
extern int lastMatchKey();
extern int saveWildKey();
extern int restoreWildKey();
extern int updateWildKey();
extern int resetWildKey();
extern int showHZfullCode();
extern int reshowHZfullCode();
extern int saveCachtoFile();
extern int loadInTab_a();
extern int hzListRelIdx();
extern int hzListAbsIdx();
extern CACH1 * getCach1Idx();
extern CACH2 * getCach2Idx();
extern int updateCach();
extern int initLXCZ();

/*** local variables ***/

static HZinputTable hzInputTables[MAX_HZTBL];	/* HZ input tables */
static HZinputType hzInputType[MAX_HZTBL];	/* HZ type tables */
static int (*hzInputFilters[MAX_HZTBL])();	/* input filter functions */
HZinputTable *cHZtbl;				/* current HZ input table */
HZinputType *cHZtyp;			/* current HZ type table */
int (*HZfilterInput)();				/* current filter function */

/* local static function declarations here */
static int HZLoadInputTable();
static int hzTableFilter();
static int hzASCIIfilter();
static int hzICfilter();
static int hzQWfilter();
static void initBufAndChoice();
static int makeChoice();
static void moveChoiceR();
static void moveChoiceL();
static void fillDpyChoices();
static void saveInputHistory();
static void restoreInputHistory();
static void clearInputHistory();
static void setMaxChoices();

struct pdHZinput {
    char *name;			/* name of input method */
    char *prompt;		/* prompt of the input method */
    unsigned char multiChoice;	/* Need multiple choices? */
       /* four system defaults added by zxk */ 
    unsigned char autoReload;	/* Need Automatic Reload? */
    unsigned char autoStudy;    /* Need automatic Study? */
    unsigned char lxDic;        /* Need LX Dictionary? */
    unsigned char fullCode;     /* Need show full encode? */
    int (*filter)();		/* HZ filter routine */
};

/* GB predefined */
static struct pdHZinput GBpredefine[] = {
  {
/* 0: The first one must be ASCII -- ISO 8859-1 */
	"ASCII",
	"\323\242\316\304\312\344\310\353 (ASCII input)",
	    /* yin wen shu ru (English Input) */
        FALSE,
        FALSE, 	FALSE, FALSE, FALSE, 
	hzASCIIfilter,
  },
  {
/* 1: IC -- Internal Code, encoding independent */
	"IC",
	"\272\272\327\326\312\344\310\353\241\313\304\332\302\353\241\313",
	    /* han zi shu ru (Chinese Input) :: nei ma (internal code) :: */
	FALSE,
        FALSE, 	FALSE, FALSE, FALSE, 
	hzICfilter,
  },
  {
/* 2: QW -- Position, for GB coding */
	"QW",
	"\272\272\327\326\312\344\310\353\241\313\307\370\316\273\241\313",
	    /* han zi shu ru (Chinese Input) :: qu wei (position) :: */
	FALSE,
        FALSE, 	FALSE, FALSE, FALSE, 
	hzQWfilter,
  },
  {
/* last: null-pad */
	NULL, NULL, FALSE, FALSE, FALSE, FALSE, FALSE, hzTableFilter,
  },
};

/* BIG5 predefined */
static struct pdHZinput BIG5predefine[] = {
  {
/* 0: The first one must be ASCII -- ISO 8859-1 */
	"ASCII",
	"\255\136\244\345\277\351\244\112 (ASCII input)",
	    /* yin wen shu ru (English Input) */
        FALSE,
        FALSE, 	FALSE, FALSE, FALSE, 
	hzASCIIfilter,
  },
  {
/* 1: IC -- Internal Code, encoding independent */
	"IC",
	"\272\176\246\162\277\351\244\112\241\107\244\272\275\130\241\107",
	    /* han zi shu ru (Chinese Input) : nei ma (internal code) : */
	FALSE,
        FALSE, 	FALSE, FALSE, FALSE, 
	hzICfilter,
  },
  {
/* 2: null-pad */
	NULL, NULL, FALSE, FALSE, FALSE, FALSE, FALSE, hzTableFilter,
  },
};

static char *hzInputMethodNames[MAX_HZTBL];
static int numHZinputs;		/* number of HZ input methods loaded */

static Char hzinbuf[MAX_INBUF];	/* input buffer for conversion to HZ */
static int hzinbufCount;	/* input buffer character count */
static trieNode *hzinbufTriePtr[MAX_INBUF];
		/* pointers to Trie Nodes for each char in hzinbuf[] */
static trieNode *cTrieNode;	/* current Trie Node */
/* static XChar2b *hzChoicePtr;	 HZ choices to be selected */
/* static int numChoice;		 number of choices on the screen */
/* static int numChoiceL;		 number of choices to the left */
static int numChoiceR;		/* number of choices to the right */
                      /* the following static variables added by zxk */
static int hzChoiceIdx;
static int numHZinf;
static int cHZinfCount;
typedef struct _HZINFORMATION {
     unsigned char relIdx;
     unsigned char numItem;
    } HZINF;
static HZINF *hzinfP=NULL;

/* saved input state variable */
static int inputSaved = 0;
static Char save_hzinbuf[MAX_INBUF];
static int save_hzinbufCount;
/* static XChar2b *save_hzChoicePtr; */
static int  save_numChoiceR /* save_numChoice, save_numChoiceL */;
       /* the following static variable added by zxk */
static int save_hzChoiceIdx;
static int save_cHZinfCount;
static int save_numItem;
#define IsCachState ( (hzinbufCount < 3) && (cHZinfCount == 0) &&\
                      (cHZtbl->autoStudy) )
#define STUDY_BY_ITSELF 64

/*
 * We do not display hzinbuf[] directly on the input area. Instead,
 * each key in the input buffer for HZ conversion is replaced by its
 * "keyprompt" string and is put into dpyInbuf[]. We show
 * dpyInbuf[] directly on the input area.
 */
#define	MAX_DPYINBUF	( MAX_INBUF * MAX_KEYPROMPT )
Char dpyInbuf[MAX_DPYINBUF];	/* user input keys to be displayed */
int dpyInbufLen;		/* length of dpyInbuf[] */

/*
 * We put HZ choices from hzChoicePtr into dpyChoices[] in the form of
 * " l.XX", l is the label of the choice, and XX is the HZ choice.
 */
/* #define	MAX_DPYCHOICE	( MAX_CHOICE * 5  + 2 ) */
#define	MAX_DPYCHOICE	256      /* modified by zxk */
Char dpyChoices[MAX_DPYCHOICE];	/* input choices to be displayed */
int dpyChoicesLen;		/* length of dpyChoices[] */

/******************************************************************/

/* ARGSUSED */
void HandleSwitchHZMode(w, event, params, nparams)
    Widget w;
    XEvent *event;
    String *params;
    Cardinal *nparams;
{
    TScreen *screen = &term->screen;

    if (*nparams == 1) {
	SetHZMode (screen, params[0], True);	/* redraw */
    }
}

SetHZMode (screen, name, doredraw)
    TScreen *screen;
    String name;
    Bool doredraw;
{
    register int i;

    for (i = 0; i < numHZinputs; i++) {
	if (! strcmp(name, hzInputMethodNames[i])) {
	    /* already loaded */
	    break;
	}
    }
    if (i == numHZinputs) {	/* not loaded */
	if (numHZinputs == MAX_HZTBL) {
	    /* too many input methods */
	    HZi_ShowMesg (screen, "Too many input methods");
	    return;
	} else {
	    if (HZLoadInputTable (screen, name, &(hzInputTables[i]), 
               &hzInputType[i])) {
		/* non-zero return: loading unsuccessfully */
		return;
	    }
	    hzInputMethodNames[i] = XtNewString(name);
	    hzInputFilters[i] = hzTableFilter;
	    numHZinputs++ ;
	}
    }
         /* added by zxk for supporting automatic reload input table */
    else if( hzInputTables[i].autoReload == 1 ) 
    {  	  free (hzInputTables[i].trieList);
	  free (hzInputTables[i].hzList);
          if( hzInputTables[i].autoStudy ) {
            free (hzInputTables[i].cachTab1);
            free (hzInputTables[i].cachTab2);
           } 
            if (HZLoadInputTable (screen, name, &(hzInputTables[i]), 
                &(hzInputType[i])) ) {
		/* non-zero return: loading unsuccessfully */
		return;
	    }
        }

    switchHZmodeByNum (screen, i);
    if (doredraw)
	RefreshHZInputArea (screen);
}

void switchHZmodeByNum (screen, num)
    TScreen *screen;
    int num;
{
    if (num >= numHZinputs)
	return;

    /* switch cxterm HZ state to mode i */
    screen->hzIwin.hzMode = num;	/* change hzIwin.hzMode to i */
    cHZtbl = &(hzInputTables[num]);
    cHZtyp = &(hzInputType[num]);
    HZfilterInput = hzInputFilters[num];
    setMaxChoices (screen, cHZtbl->maxchoice);	/* reset screen->hzIwin.max_rc */
    initBufAndChoice (cHZtbl);
    clearInputHistory ();
    initLXCZ(cHZtbl, cHZtyp);  /* added by zxk for initial static variables, */
                               /* which located in file lxcz.c */

    return;
}

/*
 * HZInitInputTable -- initialize the HZ Input Tables
 */
HZInitInputTable(screen)
    TScreen *screen;
{
    register struct pdHZinput *p = GBpredefine;

    if (screen->hzIwin.hzEncoding == GB_ENCODE) {
	p = GBpredefine;
    } else if (screen->hzIwin.hzEncoding == BIG5_ENCODE) {
	p = BIG5predefine;
    }

    numHZinputs = 0;
    while (p->name) {
	hzInputMethodNames[ numHZinputs ] = p->name;
	hzInputTables[ numHZinputs ].builtin = TRUE;
	strcpy (hzInputTables[ numHZinputs ].prompt, p->prompt);
	hzInputTables[ numHZinputs ].lenPrompt = strlen (p->prompt);
	hzInputTables[ numHZinputs ].multiChoice = p->multiChoice;
	hzInputTables[ numHZinputs ].autoReload = p->autoReload;
	hzInputTables[ numHZinputs ].autoStudy = p->autoStudy;
	hzInputFilters[ numHZinputs ] = p->filter;

	numHZinputs++;
	p++;
    }

    switchHZmodeByNum (screen, 0);	/* initially, ASCII mode */
}

/*
 * HZClearInputTable -- clear the HZ Input Tables
 */
HZClearInputTable(screen)
    TScreen *screen;
{
    register int i;

    for (i = 0; i < numHZinputs; i++) {
	if (! hzInputTables[i].builtin) {
	    if (hzInputTables[i].trieList)
		free (hzInputTables[i].trieList);
	    if (hzInputTables[i].hzList)
		free (hzInputTables[i].hzList);
            if (hzInputTables[i].autoStudy)
            {   free(hzInputTables[i].cachTab1);
                free(hzInputTables[i].cachTab2);
              }
            if (hzInputTables[i].lxDic)
                freelxdic();
	}
    }
    numHZinputs = 0;
}

int HZParseInput (screen, nbytes, strbuf)
    TScreen *screen;
    int nbytes;
    char strbuf[];
{
    /* optimize the simply cases */
    if (nbytes == 1) {
	return ((* HZfilterInput) (screen, cHZtbl, strbuf[0], strbuf));
    } else if (nbytes == 0) {
	return (0);
    } else {
	char ostr[100];	/* match the STRBUFSIZE in input.c */
	int i, nbytesOut = 0;

	strncpy (ostr, strbuf, nbytes);
	for (i = 0; i < nbytes; i++) {
	    nbytesOut += (* HZfilterInput) (screen, cHZtbl, ostr[i],
					&strbuf[nbytesOut]);
	}
	return (nbytesOut);
    }
}

/*
 * HZLoadInputTable -- load the HZ input table
 */
static int HZLoadInputTable(screen, name, hztbl, hztyp)
    TScreen *screen;
    char *name;
    HZinputTable *hztbl;
    HZinputType *hztyp;
{
    char filename[MAXPATHLEN];
    char tmpstr[80], magic[2];
    Boolean found = FALSE; 
    FILE *file;

    if (name[0] == '/') {
	/* try root directory first */
	strcpy (filename, name);
	strcat (filename, CIT_SUFFIX);
	if (access (filename, R_OK) == 0)
        {   strcpy(hztyp->fontName,filename);
	    found = TRUE; }
    }

    /* search name in different dirs */
    if ((! found) && screen->hzIwin.it_dirs) {
        register char *dir = screen->hzIwin.it_dirs;
	register char *pfilename;

	while (*dir && (! found)) {
	    /* copy from (dir) to (filename), till ':' or end of string */
	    pfilename = filename;
	    while ((*dir != '\0') && (*dir != ':'))
		*pfilename++ = *dir++ ;
	    *pfilename = '\0';

	    strcat (filename, "/");
	    strcat (filename, name);
	    strcat (filename, CIT_SUFFIX);
	    if (access (filename, R_OK) == 0)
            {   strcpy(hztyp->fontName,filename);
		found = TRUE; 
               }

	    if (*dir == ':')
		dir++ ;	/* skip this ':', ready for next component dir */
	}
    }

    if ((! found) && (name[0] != '/')) {
	/* try current directory */
	strcpy (filename, name);
	strcat (filename, CIT_SUFFIX);
	if (access (filename, R_OK) == 0)
        {   strcpy(hztyp->fontName,filename);
	    found = TRUE; }
    }

    if (! found) {
	sprintf (tmpstr, "Unable to load input table for %s", name);
	HZi_ShowMesg (screen, tmpstr);
	return(-1);
    }
    file = fopen (filename, "r");
    if (! file) {
	sprintf (tmpstr, "Unable to load input table for %s", name);
	HZi_ShowMesg (screen, tmpstr);
	return(-1);
    }
    HZi_ShowMesg (screen, "Loading input table ...");
    (void) fread (magic, 2, 1, file);
    if (strncmp (magic, MAGIC_CIT, 2) != 0) {
	sprintf (tmpstr, "Not a legal %s file for %s", CIT_SUFFIX, name);
	HZi_ShowMesg (screen, tmpstr);
        fclose(file);
	return(-1);
    }
    if (fread(hztbl, sizeof(HZinputTable), 1, file) == 0) {
	sprintf(tmpstr, "Error in loading input table for %s", name);
	HZi_ShowMesg (screen, tmpstr);
        fclose(file);
	return(-1);
    }
    if (hztbl->encode != screen->hzIwin.hzEncoding) {
	sprintf(tmpstr, "%s is not a %s encoding input method", name,
		term->misc.hz_encoding);
	HZi_ShowMesg (screen, tmpstr);
        fclose(file);
	return(-1);
    }
    hztbl->trieList = (trieNode *) calloc (hztbl->sizeTrieList,
					   sizeof(trieNode));
    hztbl->hzList = (unsigned char *)calloc(hztbl->sizeHZList+1,sizeof(char));
    if ((! hztbl->hzList) || (! hztbl->trieList)) {
	    sprintf(tmpstr, "No memory to load input table for %s", name);
	    HZi_ShowMesg (screen, tmpstr);
	    if (hztbl->trieList)  free (hztbl->trieList);
	    if (hztbl->hzList)  free (hztbl->hzList);
            fclose(file);
	    return(-1);
    }
    if ((fread(hztbl->trieList, sizeof(trieNode), hztbl->sizeTrieList, file) 
		!= hztbl->sizeTrieList) ||
	(fread(hztbl->hzList, sizeof(char), hztbl->sizeHZList, file)
		!= hztbl->sizeHZList)) {
	    sprintf(tmpstr, "Error in loading input table for %s", name);
	    HZi_ShowMesg (screen, tmpstr);
	    free (hztbl->trieList);
	    free (hztbl->hzList);
            fclose(file);
	    return(-1);
    }
           /* added by zxk */
    hztbl->hzList[hztbl->sizeHZList] = '\0';
    if( filename[0] == '/' )
    {  register char *p=&filename[strlen(filename)];
       while( *p != '/' )  
              p--;
       *p = '\0';
       if( loadInTab_a(file, filename, name, hztbl, hztyp, screen ) == -1 )
          return( -1 );
      }
    else
    {  if( loadInTab_a(file, filename, name, hztbl, hztyp, screen ) == -1 )
          return( -1 );
      }
    fclose(file);

    if( hzinfP == NULL )
    {  if( (hzinfP=(HZINF *)calloc(2,sizeof(HZINF))) == NULL )
        {    sprintf(tmpstr, "Error in alloc HZINF");
	     HZi_ShowMesg (screen, tmpstr);
             return(-1);
           }
       numHZinf = 2;
       hzinfP[0].relIdx = 0;
       hzinfP[0].numItem = 0;
      }
    return(0);
}

/**************************** Filters ***********************************/
/* Function Template:							*/
/*									*/
/* int hzXXXfilter (screen, hztbl, ch, strbuf)				*/
/*    TScreen *screen;							*/
/*    HZinputTable *hztbl;						*/
/*    unsigned char ch;							*/
/*    char strbuf[];							*/
/************************************************************************/

/*
 * hzTableFilter -- filter the input character using Input Table.
 *		Converted string are stored in strbuf, the length is returned.
 *		Return 0 if the character is eaten here.
 */
int
normalKeyProcess(screen, hztbl, ch, strbuf)
    TScreen *screen;
    HZinputTable *hztbl;
    unsigned char ch;
    char strbuf[];
{
    register int i;
    register trieNode *tnptr;

    tnptr = &(hztbl->trieList[cTrieNode->tn_nextKeys]);
	/* YGZ-XXX: it may be faster if we do binary search. */
    for (i = 0; i < cTrieNode->tn_numNextKeys; i++, tnptr++) {
       int j;
       if (ch != tnptr->tn_key) 		continue;
	  	  /* then we got it */
       if( (! hztbl->multiChoice) &&
  	  (tnptr->tn_numNextKeys == 0) &&
	  (tnptr->tn_numHZchoice == 1)) {
          int len, j, k;
		    /* multi choice is not necessary,	*/
		    /* no more input key is needed,	*/
		    /* and only one possible choice -- that's it! */
          len = ((unsigned)(hztbl->hzList[tnptr->tn_hzidx])) + 
                tnptr->tn_hzidx + 1;
   	  for( k=0,j=tnptr->tn_hzidx+1; j<len; j++ ) 
	      strbuf[k++] = hztbl->hzList[j];
          j = hzinbufCount;
	  initBufAndChoice (hztbl); /* must reset before Clear */
          HZi_ClearHzInput (screen);
          updateCach(hzinbuf, j, tnptr->tn_hzidx, hztbl, cHZtyp);
#ifdef CHKCACH
chkCach(screen, hztbl, cHZtyp, hzinbuf, tnptr->tn_hzidx);
#endif 
          resetWildKey();
          if( ++(cHZtyp->inKeyCount) == STUDY_BY_ITSELF )
          {   cHZtyp->inKeyCount = 0;
              saveCachtoFile(hztbl, cHZtyp, screen);
             }      
          if( hztbl->fullCode )
            showHZfullCode(hztbl, hzinbuf, j, tnptr->tn_hzidx, screen);
          if( hztbl->lxDic )
          {  updateGBLXDic(hztbl, strbuf);
             showlxDic(screen, hztbl, dpyInbuf, &dpyInbufLen, dpyChoices,
             &dpyChoicesLen,strbuf,(unsigned)(hztbl->hzList[tnptr->tn_hzidx]));
            }
	  return ((unsigned)(hztbl->hzList[tnptr->tn_hzidx]));
		/* "Len" bytes converted in strbuf[] */
	 }
       hzinbuf[hzinbufCount] = ch;
       hzinbufTriePtr[ hzinbufCount++ ] = cTrieNode;
       cTrieNode = tnptr;
       for( j = 0; j < hztbl->keyprompt[ch].ptlen; j++) {
   	  dpyInbuf[ dpyInbufLen++ ] = hztbl->keyprompt[ch].prompt[j];
	 }
       HZi_ReshowUserInput (screen);
       if( makeChoice (hztbl, screen->max_col) == 0)
           return( -2 ); 
       HZi_ShowInputChoices (screen);
       return (0);
     }
   return(-1);
}

static int hzTableFilter(screen, hztbl, ch, strbuf)
    TScreen *screen;
    HZinputTable *hztbl;
    unsigned char ch;
    char strbuf[];
{
    if ((ch & 0x80) || (hztbl->keytype[ch] == HZ_KEY_INVALID))
       return (1);	/* not a valid key for this input, pass out */
     
    if (hzinbufCount==0 && getLXState() &&
         (hztbl->keytype[ch] & HZ_KEY_SELECTION_MASK) ) {
        int len, sel;
                      /* process LX, zxk */
        if( (len=lxselect(ch, hztbl, dpyChoices, strbuf)) == 0 )
             return( 0 );  
        updateGBLXDic(hztbl, strbuf);
	initBufAndChoice (hztbl);
        resetWildKey();
	HZi_ClearHzInput (screen);
        if( hztbl->fullCode )
        {   register int pos=hzListAbsIdx(hztbl, strbuf);
            if( pos >= 0 )
              showHZfullCode(hztbl, strbuf, 0, pos, screen);
           }
        if( hztbl->lxDic )
            showlxDic(screen, hztbl, dpyInbuf, &dpyInbufLen,
                dpyChoices, &dpyChoicesLen, strbuf, len);
        return(len); 
      }
    else if( getLXState() )            /* reset LX, zxk */
       {   resetLX();
	   initBufAndChoice (hztbl);
	   HZi_ClearHzInput (screen);
          }

    if (hztbl->keytype[ch] & HZ_KEY_INPUT_MASK) 
    {  int ret_f;
       if (hzinbufCount >= MAX_INBUF)
	{    Bell();
	    return (0);
	   }
             /* process Wild Key, zxk */
       if( WildKeyState() )
          ret_f = wildKeyInput(ch, hztbl,cTrieNode,hzinbuf,&hzinbufCount);
       if( ret_f != -1 && WildKeyState() == 2 )
        { saveWildKey();
          while( (ret_f=wildKeyProcess(screen, hztbl, dpyInbuf, &dpyInbufLen,
             hzinbuf, hzinbufCount, hzinbufTriePtr, &cTrieNode)) == 0 )
          {  ret_f = wildKeyNext(hztbl, hzinbuf, &hzinbufCount); 
             if( ret_f == -1 )   
             {   hzinbufCount--; 
                 restoreWildKey(hzinbufCount, hzinbuf);
                 break;
                }
            }
          if( ret_f == 1 )
          { HZi_ReshowUserInput (screen);
            if( makeChoice (hztbl, screen->max_col) == 0)
                return( -2 ); 
            HZi_ShowInputChoices (screen);
            return(0);
           }
         }  
                    /* normal key process */
     else if( WildKeyState() < 2 )   
      { if( (ret_f=normalKeyProcess(screen, hztbl, ch, strbuf)) >=0 )
            return(ret_f);
        else if( ret_f == -2 )  return(0);
       }
     }
    if (hztbl->keytype[ch] & HZ_KEY_SELECTION_MASK) {
        int len, sel, absIdx;
                      /* process LX, zxk */
	if (hzinfP[cHZinfCount].numItem == 0)
	    return (1);		/* nothing to select, pass the key out */
	sel = (hztbl->keytype[ch] & HZ_KEY_SELECTION_NUM);
	if (sel >= hzinfP[cHZinfCount].numItem) {
	    Bell();
	    return (0);
	}
                    /* process hot key cache, zxk */
        if( IsCachState )   
        { if( hzinbufCount == 1 ) 
          { register CACH1 *cachP=getCach1Idx(hzinbuf, hztbl, cHZtyp);
            absIdx = hzChoiceIdx+cachP[sel]; /* for HZ aboslute idx */
           }
          else
          { register CACH2 *cachP=getCach2Idx(hzinbuf, hztbl, cHZtyp);
            absIdx = hzChoiceIdx+cachP[sel];
           }
         }
             /* normal key  which is more than 2, zxk*/ 
        else   absIdx = hzListRelIdx(sel, hzChoiceIdx, hztbl) + hzChoiceIdx;
             /* update hot key cache */
        updateCach(hzinbuf,hzinbufCount, absIdx, hztbl, cHZtyp);
#ifdef CHKCACH
chkCach(screen, hztbl,cHZtyp, hzinbuf, absIdx);
#endif 
             /* select Han Zi */
        { char *dp;
          ch = hztbl->choicelb[sel];
          if( (dp = strchr(dpyChoices, ch)) == NULL)
           {  Bell();  return 0;	} 
          dp += 1; 
          len = 0;
          while ( *dp != ' ' )
          {   strbuf[len++] = *dp++;   }	
         }
             /* process automatic study, zxk */
        if( ++(cHZtyp->inKeyCount) == STUDY_BY_ITSELF )
        {   cHZtyp->inKeyCount = 0;
               /* save new knowledge into input table file */
            saveCachtoFile(hztbl, cHZtyp, screen);
           }      
	saveInputHistory ();	/* save this HZ input */
        sel = hzinbufCount;
	initBufAndChoice (hztbl);
        resetWildKey();
	HZi_ClearHzInput (screen);
                /* display full encode of last HZ word, zxk */
        if( hztbl->fullCode )
           showHZfullCode(hztbl, hzinbuf, sel, absIdx, screen);
                /* process LX, zxk */
        if( hztbl->lxDic )
        {  updateGBLXDic(hztbl, strbuf);
           showlxDic(screen, hztbl, dpyInbuf, &dpyInbufLen,
               dpyChoices, &dpyChoicesLen, strbuf, len);
          }
       /* len bytes converted in strbuf[] */
        return(len); 
    }

    if (hztbl->keytype[ch] == HZ_KEY_BACKSPACE) {
	if (hzinbufCount == 0)
	    return (1);		/* nothing to delete, pass the key out */
	dpyInbufLen -= hztbl->keyprompt[ hzinbuf[--hzinbufCount] ].ptlen;
        updateWildKey(hzinbufCount);
	HZi_ClearHzInput (screen);	/* clear from last input key */

	cTrieNode = hzinbufTriePtr[ hzinbufCount ];
        if( makeChoice(hztbl, screen->max_col) == 0 )
            return(0);
	HZi_ShowInputChoices (screen);

        if( hztbl->fullCode )  /* reshow full encode of last HZ word, zxk */
           reshowHZfullCode(screen);
	return (0);

    } else if (hztbl->keytype[ch] == HZ_KEY_KILL) {
	if (hzinbufCount == 0)
	{  if( hztbl->fullCode ) /* reshow full encode, zxk */
               reshowHZfullCode(screen);
           return (1);		/* nothing to kill, pass the key out */
          }
	initBufAndChoice (hztbl);
        resetWildKey();
	HZi_ClearHzInput (screen);
        if( hztbl->fullCode )  /* reshow full encode of last HZ word, zxk */
           reshowHZfullCode(screen);
	return (0);

    } else if (hztbl->keytype[ch] == HZ_KEY_RIGHT) {
	if ( hzinfP[cHZinfCount].numItem == 0)
	    return (1);		/* nothing to move, pass the key out */
	if(numChoiceR == 0 || (IsCachState && 
            hzinfP[0].numItem == numChoiceR) )
	{            /* process wild code,zxk */
           if( WildKeyState() == 2 )
            { saveWildKey();
              do{
                 if (wildKeyNext(hztbl, hzinbuf, &hzinbufCount) == -1 ) 
                 {  restoreWildKey(hzinbufCount, hzinbuf);
                    Bell();  return (0);
                   }
              } while( wildKeyProcess(screen,hztbl,dpyInbuf,&dpyInbufLen,
                 hzinbuf, hzinbufCount, hzinbufTriePtr, &cTrieNode) == 0 );
              HZi_ReshowUserInput (screen);
              if( makeChoice (hztbl, screen->max_col) == 0)
                  return( -2 ); 
              HZi_ShowInputChoices (screen);
              return(0);
             }
            else
            {  Bell ();
	       return (0);		/* no more choice to the right */
	      }
         }
	moveChoiceR (hztbl, screen->max_col);
	HZi_ShowInputChoices (screen);
	return (0);

    } else if (hztbl->keytype[ch] == HZ_KEY_LEFT) {
	if (hzinfP[cHZinfCount].numItem == 0)
	    return (1);		/* nothing to move, pass the key out */
	if( cHZinfCount < 1 || ((cHZinfCount==1) && ( 
            (hzinbufCount>2) || !hztbl->autoStudy) ) ) 
        {         /* process wild code,zxk */
           if( WildKeyState() == 2 )
            { saveWildKey();
              do{
                 if (wildKeyPrev(hztbl, hzinbuf, &hzinbufCount) == -1 ) 
                 {  restoreWildKey(hzinbufCount, hzinbuf);
                    Bell();  return (0); 
                  }
              } while( wildKeyProcess(screen,hztbl,dpyInbuf,&dpyInbufLen,
                 hzinbuf, hzinbufCount, hzinbufTriePtr, &cTrieNode)==0 );
              HZi_ReshowUserInput (screen);
              if( preMakeChoice (hztbl, screen->max_col) == 0)
                  return( -2 ); 
              HZi_ShowInputChoices (screen);
              return(0);
             }
           else
           {  Bell ();
	      return (0);		/* no more choice to the left */
  	    }
         }
	moveChoiceL (hztbl, screen->max_col);
	HZi_ShowInputChoices (screen);
	return (0);

    } else if (hztbl->keytype[ch] == HZ_KEY_REPEAT) {
	if (hzinbufCount != 0)
	    return (1);		/* enable restore only when hzinbuf is empty */
	if (! inputSaved)
	    return (1);		/* nothing saved before, pass the key out */

	restoreInputHistory (hztbl);
	HZi_ReshowUserInput (screen);
	HZi_ShowInputChoices (screen);
	return (0);
    }
    /* unknown (yet valid) key */
    Bell ();
    return (0);	
}

static int hzASCIIfilter (screen, hztbl, ch, strbuf)
    TScreen *screen;
    HZinputTable *hztbl;
    unsigned char ch;
    char strbuf[];
{
    strbuf[0] = ch;
    return(1);
}

#define hex2dec(ch)			\
	(isdigit(ch) ? (ch)-'0' :	\
		(((ch)>='a') && ((ch)<='f')) ? (ch)-'a'+10 : (ch)-'A'+10)

static int hzICfilter (screen, hztbl, ch, strbuf)
    TScreen *screen;
    HZinputTable *hztbl;
    unsigned char ch;
    char strbuf[];
{
    if (isxdigit (ch)) {
	hzinbuf[ hzinbufCount++ ] = ch;
	dpyInbuf[ dpyInbufLen++ ] = ch;
	HZi_ReshowUserInput (screen);

	if (hzinbufCount == 4) {
	    strbuf[0] = hex2dec(hzinbuf[0]) * 16 + hex2dec(hzinbuf[1]);
	    strbuf[1] = hex2dec(hzinbuf[2]) * 16 + hex2dec(hzinbuf[3]);

	    hzinbufCount = 0;
	    dpyInbufLen = 0;
	    HZi_ClearHzInput (screen);
	    return (2);	/* 2 bytes converted in strbuf[] */
	} else
	    return (0);
    }

    switch (ch) {
	case '\010':	/* \b */
	case '\177':	/* DEL */
	    if (hzinbufCount == 0)
		return (1);	/* nothing to delete, pass the key out */
	    hzinbufCount-- ;
	    dpyInbufLen-- ;
	    HZi_ClearHzInput (screen);	/* clear last input key */
	    return (0);
	    break;

	case '\015':	/* \r */
	case '\025':	/* ^U */

	    if (hzinbufCount == 0)
		return (1);	/* nothing to kill, pass the key out */
	    hzinbufCount = 0;
	    dpyInbufLen = 0;
	    HZi_ClearHzInput (screen);
	    return (0);
	    break;
    }

    return (1);
}

/* GB: QW */
static int hzQWfilter (screen, hztbl, ch, strbuf)
    TScreen *screen;
    HZinputTable *hztbl;
    unsigned char ch;
    char strbuf[];
{
    if (isdigit (ch)) {
	hzinbuf[ hzinbufCount++ ] = ch;
	dpyInbuf[ dpyInbufLen++ ] = ch;
	HZi_ReshowUserInput (screen);

	if (hzinbufCount == 4) {
	    strbuf[0] = (hzinbuf[0] - '0') * 10 + (hzinbuf[1] - '0') + 0xa0;
	    strbuf[1] = (hzinbuf[2] - '0') * 10 + (hzinbuf[3] - '0') + 0xa0;

	    hzinbufCount = 0;
	    dpyInbufLen = 0;
	    HZi_ClearHzInput (screen);
	    return (2);	/* 2 bytes converted in strbuf[] */
	} else
	    return (0);
    }

    switch (ch) {
	case '\010':	/* \b */
	case '\177':	/* DEL */
	    if (hzinbufCount == 0)
		return (1);	/* nothing to delete, pass the key out */
	    hzinbufCount-- ;
	    dpyInbufLen-- ;
	    HZi_ClearHzInput (screen);	/* clear last input key */
	    return (0);
	    break;

	case '\015':	/* \r */
	case '\025':	/* ^U */

	    if (hzinbufCount == 0)
		return (1);	/* nothing to kill, pass the key out */
	    hzinbufCount = 0;
	    dpyInbufLen = 0;
	    HZi_ClearHzInput (screen);
	    return (0);
	    break;
    }

    return (1);
}


static void initBufAndChoice (hztbl)
    HZinputTable *hztbl;
{
    hzinbufCount = 0;
    dpyInbufLen = 0;
    numChoiceR = 0;
    dpyChoicesLen = 0;
    cTrieNode = hztbl->trieList;
        cHZinfCount = 0;
    if( hzinfP != NULL )
    {  hzinfP[0].relIdx = 0;
       hzinfP[0].numItem = 0;
      }
    hzChoiceIdx = 0;
}

RemakeHZInput (screen)
    TScreen *screen;
{
    /* (re-) Calculate maximun number of choices on the screen */
    int oldNumChoice;
    setMaxChoices (screen, cHZtbl->maxchoice);

    if ((! cHZtbl) || cHZtbl->builtin)	/* no Choices */
	return;
    oldNumChoice = hzinfP[cHZinfCount].numItem;
    numChoiceR += oldNumChoice;
    hzinfP[cHZinfCount].numItem = findHZDispLen(screen->max_col,  
                  cTrieNode->tn_numHZchoice, cHZtbl);
    hzinfP[cHZinfCount].relIdx = hzListRelIdx(hzinfP[cHZinfCount].numItem,
               hzChoiceIdx, cHZtbl);
    if (hzinfP[cHZinfCount].numItem < oldNumChoice) {
	numChoiceR -= hzinfP[cHZinfCount].numItem;
	fillDpyChoices (cHZtbl);
    } else if ((oldNumChoice < hzinfP[cHZinfCount].numItem)
                 && (numChoiceR > 0)) {
        numChoiceR -= hzinfP[cHZinfCount].numItem;
	fillDpyChoices (cHZtbl);
    }
}

static int makeChoice (hztbl, max_col, screen)
    TScreen *screen;
    HZinputTable *hztbl;
    int max_col;		/* maximun choices on screen */
{
    char tmpstr[80];
           /* average 4 HZ in each input selcection screen */
    if( numHZinf < ((cTrieNode->tn_numHZchoice)>>2) )
    {  free(hzinfP);
       if( (hzinfP=(HZINF *)calloc((cTrieNode->tn_numHZchoice)>>2, 
            sizeof(HZINF))) == NULL ) 
       {    sprintf(tmpstr, "Error in alloc HZINF");
	    HZi_ShowMesg (screen, tmpstr);
            Bell();
            return(0);
          }
       numHZinf = (cTrieNode->tn_numHZchoice)>>2;
      }
    numChoiceR = cTrieNode->tn_numHZchoice;
    hzChoiceIdx = cTrieNode->tn_hzidx;
    if( hzinbufCount < 3 && hztbl->autoStudy ) {
        cHZinfCount = 0;
        hzinfP[0].numItem = 
            findHZDispLen(max_col, cTrieNode->tn_numHZchoice, hztbl);
        hzinfP[0].relIdx = 0;
      }
    else {
        cHZinfCount = 1;
        hzinfP[1].numItem = 
            findHZDispLen(max_col, cTrieNode->tn_numHZchoice, hztbl);
        hzinfP[1].relIdx = hzListRelIdx(hzinfP[1].numItem, hzChoiceIdx, hztbl);
        numChoiceR -= (unsigned)hzinfP[1].numItem;
       }
    hzChoiceIdx = cTrieNode->tn_hzidx;
    fillDpyChoices (hztbl);
    return(1);
}

static int       /* this function written by zxk */
preMakeChoice (hztbl, max_col, screen)
    TScreen *screen;
    HZinputTable *hztbl;
    int max_col;		/* maximun choices on screen */
{
    char tmpstr[80];
           /* average 4 HZ in each input selcection screen */
    if( numHZinf < ((cTrieNode->tn_numHZchoice)>>2) )
    {  free(hzinfP);
       if( (hzinfP=(HZINF *)calloc((cTrieNode->tn_numHZchoice)>>2, 
            sizeof(HZINF))) == NULL ) 
       {    sprintf(tmpstr, "Error in alloc HZINF");
	    HZi_ShowMesg (screen, tmpstr);
            Bell();
            return(0);
          }
       numHZinf = (cTrieNode->tn_numHZchoice)>>2;
      }
    numChoiceR = cTrieNode->tn_numHZchoice;
    hzChoiceIdx = cTrieNode->tn_hzidx;
    if( hzinbufCount < 3 && hztbl->autoStudy ) {
        cHZinfCount = 0;
        hzinfP[0].numItem = 
            findHZDispLen(max_col, cTrieNode->tn_numHZchoice, hztbl);
        hzinfP[0].relIdx = 0;
      }
    else 
    {   cHZinfCount = 0;
        hzinfP[0].numItem = 0;
        hzinfP[0].relIdx = 0;
       }
    while( numChoiceR )
    {   if( cHZinfCount ) hzChoiceIdx += hzinfP[cHZinfCount].relIdx;
        cHZinfCount++;
        hzinfP[cHZinfCount].numItem = 
            findHZDispLen(max_col, cTrieNode->tn_numHZchoice, hztbl);
        hzinfP[cHZinfCount].relIdx = 
            hzListRelIdx(hzinfP[cHZinfCount].numItem, hzChoiceIdx, hztbl);
        numChoiceR -= (unsigned)hzinfP[cHZinfCount].numItem;
       }
    fillDpyChoices (hztbl);
    return(1);
}

static void moveChoiceR (hztbl, max_col)
    HZinputTable *hztbl;
    int max_col;		/* maximun choices on screen */
{
    if( IsCachState ) {
       cHZinfCount = 1;
       hzinfP[1].numItem = 
            findHZDispLen(max_col, cTrieNode->tn_numHZchoice, hztbl);
       hzinfP[1].relIdx = hzListRelIdx(hzinfP[1].numItem, hzChoiceIdx, hztbl);
      }
    else {
       hzChoiceIdx += hzinfP[cHZinfCount].relIdx;
       cHZinfCount++;
       hzinfP[cHZinfCount].numItem = 
            findHZDispLen(max_col, cTrieNode->tn_numHZchoice, hztbl);
       hzinfP[cHZinfCount].relIdx = hzListRelIdx(hzinfP[cHZinfCount].numItem,
                                      hzChoiceIdx, hztbl);
      }
    numChoiceR -= hzinfP[cHZinfCount].numItem;
    fillDpyChoices (hztbl);
}

static void moveChoiceL (hztbl, max_col)
    HZinputTable *hztbl;
    int max_col;		/* maximun choices on screen */
{
    if( (hzinbufCount < 3) && (cHZinfCount == 1) && 
        (hztbl->autoStudy) ) {
        hzChoiceIdx = cTrieNode->tn_hzidx;
        cHZinfCount = 0;
        numChoiceR = cTrieNode->tn_numHZchoice;
        hzinfP[0].numItem = 
            findHZDispLen(max_col, cTrieNode->tn_numHZchoice, hztbl);
        hzinfP[0].relIdx = 0;
      }
    else {
        numChoiceR += hzinfP[cHZinfCount].numItem;
        cHZinfCount--; 
        hzChoiceIdx -= hzinfP[cHZinfCount].relIdx;
      }
    fillDpyChoices (hztbl);
}


static void fillDpyChoices (hztbl)
    HZinputTable *hztbl;
{
    register int i;
    register CACH1 *cach1P;
    register CACH2 *cach2P;

    dpyChoicesLen = 0;
    if( hzinbufCount<3 && hztbl->autoStudy>0 ) 
    {  if( hzinbufCount == 1 ) cach1P = getCach1Idx(hzinbuf, hztbl, cHZtyp);
       else                  cach2P = getCach2Idx(hzinbuf, hztbl,cHZtyp);
       dpyChoices[ dpyChoicesLen ] = (cHZinfCount==0) ? ' ': '<';
     }
    else
       dpyChoices[ dpyChoicesLen ] = (cHZinfCount==1) ? ' ': '<';
    if( dpyChoices[0]==' ' && WildKeyState()>1 && !firstMatchKey() )
           dpyChoices[ dpyChoicesLen ] = '$';
    dpyChoicesLen++;
    for (i = 0; i < hzinfP[cHZinfCount].numItem; i++) {
      int len, j;
	dpyChoices[ dpyChoicesLen++ ] = ' ';
	dpyChoices[ dpyChoicesLen++ ] = hztbl->choicelb[i];
        if( IsCachState )
        { if( hzinbufCount == 1 )
          {  len = ((unsigned)(hztbl->hzList[hzChoiceIdx+cach1P[i]])) +
                 hzChoiceIdx + cach1P[i] + 1;
   	     for( j=hzChoiceIdx+cach1P[i]+1; j<len; j++ ) 
                dpyChoices[ dpyChoicesLen++ ] = hztbl->hzList[j];
            }
          else
          {  len = ((unsigned)(hztbl->hzList[hzChoiceIdx+cach2P[i]])) +
                 hzChoiceIdx + cach2P[i] + 1;
   	     for( j=hzChoiceIdx+cach2P[i]+1; j<len; j++ ) 
                dpyChoices[ dpyChoicesLen++ ] = hztbl->hzList[j];
            }
         }
        else 
        { j = hzListRelIdx(i, hzChoiceIdx, hztbl) + hzChoiceIdx;
          len = ((unsigned)(hztbl->hzList[j])) + j + 1;
          j++ ;
   	  while( j<len  ) 
          {  dpyChoices[ dpyChoicesLen++ ] = hztbl->hzList[j];
             j ++;
            }
         }
   }
    dpyChoices[ dpyChoicesLen++ ] = ' '; 
    dpyChoices[ dpyChoicesLen ] = (numChoiceR==0 || (IsCachState && 
                   numChoiceR==hzinfP[cHZinfCount].numItem)) ? ' ' : '>';
    if( dpyChoices[dpyChoicesLen]==' ' && WildKeyState()>1 && !lastMatchKey() )
           dpyChoices[ dpyChoicesLen ] = '$';
    dpyChoicesLen++;
    dpyChoices[ dpyChoicesLen ] ='\0';
}
static  int   /* this function written by zxk */
findHZDispLen(max_col, hzLen, hztbl)
	int max_col, hzLen;
        HZinputTable *hztbl;
{
	register int  bc = 3, rc = 0;

        if( IsCachState )
        { register int  maxv=(numChoiceR<16) ? numChoiceR : 16;
          if( hzinbufCount == 1 ) 
          {  register CACH1 *cachP;
             cachP = getCach1Idx(hzinbuf, hztbl, cHZtyp);
	     while ( bc < max_col && rc < maxv )	
 	     {	bc += 2 + ((unsigned)(hztbl->hzList[hzChoiceIdx+cachP[rc]]));
                rc++;
               }
            }
          else      
          {  register CACH2 *cachP;
             cachP = getCach2Idx(hzinbuf, hztbl, cHZtyp);
	     while ( bc < max_col && rc < maxv )	
 	     {	bc += 2 + ((unsigned)(hztbl->hzList[hzChoiceIdx+cachP[rc]]));
                rc++;
               }
            }
          }
        else 
        {  int idx=hzChoiceIdx;
           while ( bc < max_col && rc < numChoiceR )	
 	   {	bc += 2 + ((unsigned)(hztbl->hzList[idx]));
                idx += 	((unsigned)(hztbl->hzList[idx])) + 1;
                rc++;
             }
         }
      if( bc >= max_col ) rc--;
      if( rc > hztbl->maxchoice )     rc = hztbl->maxchoice;
      return  rc;
}


struct codetbl {
	char *codestr;
	int   codeint;
} codeTbl[] = {

	"GB",		GB_ENCODE,
	"Gb",		GB_ENCODE,
	"gb",		GB_ENCODE,
	"GUOBIAO",	GB_ENCODE,
	"GuoBiao",	GB_ENCODE,
	"guobiao",	GB_ENCODE,

	"BIG5",		BIG5_ENCODE,
	"Big5",		BIG5_ENCODE,
	"big5",		BIG5_ENCODE,
	"B5",		BIG5_ENCODE,
	"b5",		BIG5_ENCODE,

	NULL,		GB_ENCODE,	/* By default, GB_ENCODE */
};

int HZencode (name)
    char *name;
{
    struct codetbl *pct = codeTbl;

    while (pct->codestr) {
	if (strcmp (name, pct->codestr) == 0)
	    return (pct->codeint);
	pct++;
    }
    return (pct->codeint);
}

static void saveInputHistory()
{
    /* make minimum effords in save */
    save_hzinbufCount = hzinbufCount;
    strncpy (save_hzinbuf, hzinbuf, hzinbufCount);
/*    save_hzChoicePtr = hzChoicePtr; */
    save_hzChoiceIdx = hzChoiceIdx;
/*    save_numChoice  = numChoice; */
    save_numChoiceR = numChoiceR;
/*    save_numChoiceL = numChoiceL; */
    save_cHZinfCount = cHZinfCount;
    save_numItem = hzinfP[cHZinfCount].numItem;
    inputSaved = 1;
}
static void restoreInputHistory(hztbl)
    HZinputTable *hztbl;
{
  register int i;

    hzinbufCount = save_hzinbufCount;
    strncpy (hzinbuf, save_hzinbuf, hzinbufCount);
/*    hzChoicePtr = save_hzChoicePtr; */
    hzChoiceIdx = save_hzChoiceIdx;
/*    numChoice  = save_numChoice; */
    numChoiceR = save_numChoiceR;
/*    numChoiceL = save_numChoiceL; */
    cHZinfCount = save_cHZinfCount;
    hzinfP[cHZinfCount].numItem = save_numItem;

    /* recompute everything */
    for (i = 0; i < hzinbufCount; i++) {
	unsigned char ch = hzinbuf[i];
	register int j, numNextKeys;

	hzinbufTriePtr[i] = cTrieNode;
	numNextKeys = cTrieNode->tn_numNextKeys;
	cTrieNode = &(hztbl->trieList[cTrieNode->tn_nextKeys]);
	for (j = 0; j < numNextKeys; j++, cTrieNode++) {
	    if (ch == cTrieNode->tn_key)
		break;
	}

	for (j = 0; j < hztbl->keyprompt[ch].ptlen; j++) {
	    dpyInbuf[ dpyInbufLen++ ] = hztbl->keyprompt[ch].prompt[j];
	}
    }

    fillDpyChoices (hztbl);
}


static void clearInputHistory()
{
    inputSaved = 0;
}

static void setMaxChoices (screen, maxchoice)
    TScreen *screen;
    int maxchoice;
{
    /* Set the maximun number of choices on the screen */
    /* from the width of the screen and the maximun choices allowed */
    /* max_col + 1 - 3: 3 means a leading "<" and ending " >" */
    screen->hzIwin.max_rc = (screen->max_col - 2) / 5;
    if (screen->hzIwin.max_rc > maxchoice)
	screen->hzIwin.max_rc = maxchoice;
}

initBufagain (hztbl)
    HZinputTable *hztbl;
{
    numChoiceR = 0;
    dpyChoicesLen = 0;
        cHZinfCount = 0;
    if( hzinfP != NULL )
    {  hzinfP[0].relIdx = 0;
       hzinfP[0].numItem = 0;
       hzinfP[1].relIdx = 0;
       hzinfP[1].numItem = 0;
      }
    hzChoiceIdx = 0;
}

#endif /* HANZI */
